unit main;

{$mode objfpc}{$H+}

{$DEFINE FLOAT_SAMPLES}     // Normalize sample values as floats to 1

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, Types,
  StdCtrls, ExtCtrls, ComCtrls, ShellCtrls, Grids, Buttons,
  TAChartUtils, TAGraph, TACustomSource, TASources, TASeries,
  TANavigation, TATransformations,
  wavPlay, wavRead;

type
  { TMainForm }

  TMainForm = class(TForm)
    btnPlay: TBitBtn;
    Chart: TChart;
    ChartAxisTransformations1: TChartAxisTransformations;
    Grid: TDrawGrid;
    Images: TImageList;
    LinearAxisTransform: TLinearAxisTransform;
    ChartNavScrollBar: TChartNavScrollBar;
    HeaderMemo: TMemo;
    PageControl1: TPageControl;
    Panel1: TPanel;
    Panel2: TPanel;
    Panel3: TPanel;
    ShellListView: TShellListView;
    ShellTreeView: TShellTreeView;
    ShellTreeListSplitter: TSplitter;
    ShellSplitter: TSplitter;
    Splitter1: TSplitter;
    pgTable: TTabSheet;
    pgPlot: TTabSheet;
    procedure btnPlayClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure GridDrawCell(Sender: TObject; ACol, ARow: Integer; ARect: TRect;
      AState: TGridDrawState);
    procedure GridPrepareCanvas(Sender: TObject; ACol, ARow: Integer;
      AState: TGridDrawState);
    procedure ShellListViewClick(Sender: TObject);
    procedure ShellListViewFileAdded(Sender: TObject; Item: TListItem);
    procedure ShellTreeViewGetImageIndex(Sender: TObject; Node: TTreeNode);
    procedure ShellTreeViewGetSelectedIndex(Sender: TObject; Node: TTreeNode);
  private
    FWavHeader: TWavFmtSubchunk;
    {$IFDEF FLOAT_SAMPLES}
    FData: array of array of double;
    {$ELSE}
    FDataI: array of array of Int64;
    {$ENDIF}
    FSampleFormat: TWavSampleFormat;
    FFileName: String;
    FPlayer: TPlaySound;
    FPlaying: Boolean;
    procedure GetChartDataItem(ASource: TUserDefinedChartSource;
      AIndex: Integer; var AItem: TChartDataItem);
    procedure GetData(AWavFile: TWavFile);
    procedure PrepareChart(AWavFile: TWavFile);
    procedure PrepareGrid(AWavFile: TWavFile);
    procedure ShowHeader(AWavFile: TWavFile);

  public
    procedure LoadWav(AFileName: String);

  end;

var
  MainForm: TMainForm;

implementation

{$R *.lfm}

{ TMainForm }

procedure TMainForm.btnPlayClick(Sender: TObject);
begin
  if not FPlaying then begin
    if FFileName = '' then
      exit;
    FPlayer.SoundFile := FFileName;
    FPlayer.Execute;
    FPlaying := true;
    btnPlay.Caption := 'Stop';
  end else begin
    FPlayer.StopSound;
    FPlaying := false;
    btnPlay.Caption := 'Play';
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  FPlayer := TPlaySound.Create(self);
  FPlayer.PlayStyle := psAsync;
  Grid.ColWidths[0] := Grid.Canvas.TextWidth('999999999');
  ShellTreeview.Path := Application.Location;
end;

procedure TMainForm.GridDrawCell(Sender: TObject; aCol, aRow: Integer;
  aRect: TRect; aState: TGridDrawState);
var
  s: String;
  idx: Integer;
  value: Double;
begin
  s := '';
  if ARow = 0 then
    case ACol of
      0: s := 'Index';
      1: s := 'Time (s)';
    else s := 'Channel ' + IntToStr(ACol - 1);
    end
  else begin
    idx := (ARow - 1) * FWavHeader.NumChannels + (ACol - 2);
    case ACol of
      0: s := IntToStr(idx + 2);
      1: s := FormatDateTime('nn:ss.zzz', (idx + 2) / FWavHeader.SampleRate / (24*60*60));
      else
         if idx < Length(FData) then begin
           {$IFDEF FLOAT_SAMPLES}
           s := FormatFloat('0.00000', FData[idx, ACol - 2]);
           {$ELSE}
           s := IntToSTr(FDataI[idx, ACol - 2]);
           {$ENDIF}
         end;
    end;
  end;
  InflateRect(ARect, -varCellpadding, -varCellPadding);
  Grid.Canvas.TextRect(ARect, ARect.Left, ARect.Top, s);
end;

procedure TMainForm.GridPrepareCanvas(sender: TObject; aCol, aRow: Integer;
  aState: TGridDrawState);
var
  ts: TTextStyle;
begin
  ts := Grid.Canvas.TextStyle;
  if ARow = 0 then
    ts.Alignment := taCenter
  else
    ts.Alignment := taRightJustify;
  Grid.Canvas.TextStyle := ts;
end;

procedure TMainForm.GetChartDataItem(ASource: TUserDefinedChartSource;
  AIndex: Integer; var AItem: TChartDataItem);
begin
  AItem.X := AIndex;
  {$IFDEF FLOAT_SAMPLES}
  AItem.Y := FData[AIndex, ASource.Tag];
  {$ELSE}
  AItem.Y := FDataI[AIndex, ASource.TAg];
  {$ENDIF}
end;

procedure TMainForm.GetData(AWavFile: TWavFile);
var
  idx, ch: Integer;
begin
  FWavHeader := AWavFile.Header;
  FSampleFormat := AWavFile.SampleFormat;

  {$IFDEF FLOAT_SAMPLES}
  SetLength(FData, AWavFile.NumSamplesPerChannel, AWavFile.Header.NumChannels);
  for idx := 0 to High(FData) do
    for ch := 0 to AWavFile.Header.NumChannels-1 do
      FData[idx, ch] := AWavFile.ValueAsFloat[idx, ch];
  {$ELSE}
  SetLength(FDataI, AWavFile.NumSamplesPerChannel, AWavFile.Header.NumChannels);
  for idx := 0 to High(FDataI) do
    for ch := 0 to AWavFile.Header.NumChannels-1 do
      FDataI[idx, ch] := AWavFile.ValueAsInt[idx, ch];
 {$ENDIF}
end;

procedure TMainForm.LoadWav(AFileName: String);
var
  stream: TMemoryStream;
  msg: String;
  wav: TWavFile;
  ok: Boolean;
  crs: TCursor;
begin
  FFileName := AFileName;
  wav := TWavFile.Create;
  try
    stream := TMemoryStream.Create;
    try
      stream.LoadFromFile(AFileName);
      ok := wav.LoadFromStream(stream, msg);
    finally
      stream.Free;
    end;
    if ok then begin
      crs := Screen.Cursor;
      try
        Screen.Cursor := crHourglass;
        GetData(wav);
        ShowHeader(wav);
        PrepareChart(wav);
        PrepareGrid(wav);
      finally
        Screen.Cursor := crs;
      end;
    end else
      MessageDlg(msg, mtError, [mbOK], 0);
  finally
    wav.Free;
  end;
end;

procedure TMainForm.PrepareChart(AWavFile: TWavFile);
const
  COLORS: array[0..7] of TColor = (clRed, clBlue, clGreen, clBlack, clOlive, clTeal, clPurple, clMaroon);
var
  source: TUserDefinedChartSource;
  ser: TLineSeries;
  npts: Integer;
  i, ch: integer;
  ext: TDoubleRect;
begin
  // Remove series and sources from previous run
  for i := Chart.SeriesCount-1 downto 0 do begin
    ser := Chart.Series[i] as TLineSeries;
    if ser.Source <> nil then ser.Source.Free;
  end;
  Chart.ClearSeries;
  Chart.ZoomFull;

  // Create source and series
  npts := AWavFile.DataSize div AWavFile.Header.BlockAlign;
  for ch := 1 to AWavFile.Header.NumChannels do begin
    source := TUserDefinedChartSource.Create(self);
    source.PointsNumber := npts;
    source.Tag := ch - 1;
    source.Sorted := true;
    source.OnGetChartDataItem := @GetChartDataItem;
    source.Reset;

    ser := TLineSeries.Create(Chart);
    ser.Source := source;
    ser.SeriesColor := COLORS[Chart.SeriesCount mod Length(COLORS)];
    ser.Title := 'Channel ' + IntToStr(ch);
    Chart.AddSeries(ser);
  end;
  Chart.Legend.Visible := Chart.SeriesCount > 1;
  LinearAxisTransform.Scale := AWavFile.Header.SampleRate;

  Chart.Prepare;
  ext := Chart.LogicalExtent;
  ext.b.x := ext.a.x + AWavFile.Header.SampleRate / AWavFile.Header.NumChannels / 100;  // zoom to 10 ms
  Chart.LogicalExtent := ext;

  ChartNavScrollbar.Max := TLineSeries(Chart.Series[0]).Count;
  ChartNavScrollbar.Position := 0;
end;

procedure TMainForm.PrepareGrid(AWavFile: TWavFile);
begin
  with AWavFile do begin
    Grid.ColCount := Header.NumChannels + 2;
    Grid.RowCount := NumSamplesPerChannel + 1; //DataSize * 8 div Header.BitsPerSample div Header.NumChannels;
  end;
  Grid.ColWidths[0] := Grid.Canvas.TextWidth(IntToStr(Grid.RowCount)+ '    ') + 2*varCellPadding;
  Grid.Invalidate;
end;

procedure TMainForm.ShellListViewClick(Sender: TObject);
var
  fn: String;
begin
  if ShellListView.ItemFocused <> nil then
  begin
    fn := ShellListView.GetPathFromItem(ShellListView.ItemFocused);
    LoadWav(fn);
  end;
end;

procedure TMainForm.ShellListViewFileAdded(Sender: TObject; Item: TListItem);
begin
  Item.ImageIndex := 3;
end;

procedure TMainForm.ShellTreeViewGetImageIndex(Sender: TObject; Node: TTreeNode);
begin
  if Node.Level = 0 then
    Node.ImageIndex := 0
  else if Node.Selected then
    Node.ImageIndex := 2
  else
    Node.ImageIndex := 1;
end;

procedure TMainForm.ShellTreeViewGetSelectedIndex(Sender: TObject; Node: TTreeNode
  );
begin
  if Node.Level = 0 then
    Node.SelectedIndex := 0
  else if Node.Selected then
    Node.SelectedIndex := 2
  else
    Node.SelectedIndex := 1;
  BtnPlay.Enabled := true;
end;

procedure TMainForm.ShowHeader(AWavFile: TWavFile);

  procedure AddToMemo(ACaption, AValue: String);
  const
    CAPTION_SIZE = 32;
  var
    s: String;
  begin
    s := Format('%*s : %s', [CAPTION_SIZE, ACaption, AValue]);
    HeaderMemo.Lines.Add(s);
  end;

var
  s: String;
  cnv: TControlCanvas;
begin
  HeaderMemo.Lines.Clear;

  AddToMemo('File', ExtractFileName(FFileName));

  // https://de.wikipedia.org/wiki/RIFF_WAVE
  case AWavFile.Header.AudioFormat of
    $0001: s := 'PCM, uncompressed';
    $0002: s := 'MS ADPCM';
    $0003: s := 'IEEE FLOAT';
    $0005: s := 'IBM CVSD';
    $0006: s := 'ALAW';
    $0007: s := 'MULAW';
    $0010: s := 'OKI ADPCM';
    $0011: s := 'DVI/IMA ADPCM';
    $0012: s := 'MEDIASPACE ADPCM';
    $0013: s := 'SIERRA ADPCM';
    $0014: s := 'G723 ADPCM';
    $0015: s := 'DIGISTD';
    $0016: s := 'DIGIFIX';
    $0017: s := 'DIALOGIC OKI ADPCM';
    $0020: s := 'YAMAHA ADPCM';
    $0021: s := 'SONARC';
    $0022: s := 'DSPGROUP TRUESPEECH';
    $0023: s := 'ECHOSC1';
    $0024: s := 'AUDIOFILE AF36';
    $0025: s := 'APTX';
    $0026: s := 'AUDIOFILE AF10';
    $0030: s := 'DOLBY AC2';
    $0031: s := 'GSM610';
    $0033: s := 'ANTEX ADPCME';
    $0034: s := 'CONTROL RES VQLPC';
    $0035: s := 'CONTROL RES VQLPC';
    $0036: s := 'DIGIADPCM';
    $0037: s := 'CONTROL RES CR10';
    $0038: s := 'NMS VBXADPCM';
    $0039: s := 'CS IMAADPCM (Roland RDAC';
    $0040: s := 'G721 ADPCM';
    $0050: s := 'MPEG-1 Layer I, II';
    $0055: s := 'MPEG-1 Layer III (MP3)';
    $0069: s := 'Xbox ADPCM';
    $0200: s := 'CREATIVE ADPCM';
    $0202: s := 'CREATIVE FASTSPEECH8';
    $0203: s := 'CREATIVE FASTSPEECH10';
    $0300: s := 'FM TOWNS SND';
    $1000: s := 'OLIGSM';
    $1001: s := 'OLIADPCM';
    $1002: s := 'OLICELP';
    $1003: s := 'OLISBC';
    $1004: s := 'OLIOPR';
    else   s := '(unknown)';
  end;

  AddToMemo('Audio format', Format('$%0:.4x (%0:d = %1:s)', [AWavFile.Header.AudioFormat, s]));
  AddToMemo('Channels', IntToStr(AWavFile.Header.NumChannels));
  AddToMemo('SampleRate (samples/s)', IntToStr(AWavFile.Header.SampleRate));
  AddToMemo('Byte rate (bytes/s)', IntToStr(AWavFile.Header.ByteRate));
  AddToMemo('Block align', IntToStr(AWavFile.Header.BlockAlign));
  AddToMemo('Bits per sample', IntToStr(AWavFile.Header.BitsPerSample));
  AddToMemo('Size of data block (bytes)', IntToStr(AWavFile.DataSize));

  // Calculate number of samples (1 sample per channel)
  with AWavFile do begin
    HeaderMemo.Lines.Add('');
    AddToMemo('Total number of samples', IntToStr(DataSize * 8 div Header.BitsPerSample));
    AddToMemo('Number of samples per channel', IntToStr(DataSize * 8 div Header.BitsPerSample div Header.NumChannels));
    AddToMemo('Number of samples per channel', IntToStr(NumSamplesPerChannel));
    AddToMemo('Duration (s)', FormatFloat('0.0', DataSize / Header.ByteRate));
  end;

  cnv := TControlCanvas.Create;
  try
    cnv.Control := HeaderMemo;
    cnv.Font.Assign(HeaderMemo.Font);
    HeaderMemo.Height := cnv.TextHeight('Tg') * (HeaderMemo.Lines.Count + 1) + 8;
  finally
    cnv.Free;
  end;
end;


end.

