unit wavRead;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils;

type
  // http://soundfile.sapp.org/doc/WaveFormat/

  TWavRiffHeader = packed record
    ID: array[0..3] of ansichar;        // "RIFF" for little-endian, "RIFX" for big-endian
    ChunkSize: LongInt;                 // Filesize - 8
    RiffType: array[0..3] of ansiChar;  // "WAVE"
  end;

  TWavFmtSubchunk = packed record
    ID: array[0..3] of ansichar;  // "fmt "
    SubChunkSize: LongInt;   // 16
    AudioFormat: SmallInt;   // PCM = 1 --> uncompressed
    NumChannels: SmallInt;   // Mono = 1, Stereo = 2
    SampleRate: LongInt;     // 8000, 44100 etc.
    ByteRate: LongInt;       // SampleRate * NumChannels * BitsPerSample / 8
    BlockAlign: SmallInt;    // NumChannels * BitsPerSample / 8
    BitsPerSample: SmallInt; // 8, 16
  end;

  TWavSubchunkHeader = packed record
    ID: array[0..3] of ansichar;  // chunk identifier, e.g. "data"
    DataSize: LongInt;            // NumSamples * NumChannels * BitsPerSample / 8
                                  // or: size of following data part.
  end;

  TWavSampleFormat = (wsfInt8, wsfInt16, wsfInt24, wsfInt32, wsfInt64,
    wsfFloat32, wsfFloat64);

  TWavFile = class
  private
    FHeader: TWavFmtSubChunk;
    FChunkSize: LongInt;
    FData: Pointer;
    FDataSize: LongInt;
    FSampleFormat: TWavSampleFormat;
    function GetNumSamplesPerChannel: Integer;
    function GetSampleFormat: TWavSampleFormat;
    function GetValueAsInt(ASampleIndex, AChannelIndex: Integer): Int64;
    function GetValueAsFloat(ASampleIndex, AChannelIndex: Integer): Double;
  public
    destructor Destroy; override;
    function LoadFromStream(AStream: TStream; out AMsg: String): Boolean;
    property Data: Pointer read FData;
    property DataSize: LongInt read FDataSize;
    property Header: TWavFmtSubChunk read FHeader;
    property NumSamplesPerChannel: Integer read GetNumSamplesPerChannel;
    property SampleFormat: TWavSampleFormat read FSampleFormat;
    property ValueAsFloat[SampleIndex, ChannelIndex: Integer]: Double read GetValueAsFloat;
    property ValueAsInt[SampleIndex, ChannelIndex: Integer]: Int64 read GetValueAsInt;
  end;

  EWavFile = class(Exception);

implementation

{ TWavFile }

destructor TWavFile.Destroy;
begin
  FreeMem(FData, FDataSize);
  inherited;
end;

function TWavFile.GetNumSamplesPerChannel: Integer;
begin
  Result := FDataSize div FHeader.BlockAlign;
end;

function TWavFile.GetSampleFormat: TWavSampleFormat;
begin
  case FHeader.AudioFormat of
    1: case FHeader.BitsPerSample of
          8: Result := wsfInt8;
         16: Result := wsfInt16;
         24: Result := wsfInt24;
         32: Result := wsfInt32;
         64: Result := wsfInt64;
         else
           raise EWavFile.Create('Bits per sample not supported');
       end;
    3: case FHeader.BitsPerSample of
         32: Result := wsfFloat32;
         64: Result := wsfFloat64;
         else
           raise EWavFile.Create('Bits per sample not supported');
       end;
    else
      raise EWavFile.Create('Audio format not supported');
  end;
end;

type
  TData4 = packed array[0..3] of byte;
  PData4 = ^TData4;

function TWavFile.GetValueAsFloat(ASampleIndex, AChannelIndex: Integer): Double;
var
  idx: Integer;
  d24: TData4 = (0, 0, 0, 0);
begin
  idx := ASampleIndex * FHeader.BlockAlign + AChannelIndex * FHeader.BitsPerSample div 8;
  case FSampleFormat of
    wsfInt8:
      Result := 1.0 * PByte(FData + idx)^ / $7F;
    wsfInt16:
      Result := 1.0 * PSmallInt(FData + idx)^ / $7FFF;
    wsfInt24:
      begin
        Move((FData + idx)^, d24[1], 3);
        Result := 1.0 * (LongInt(d24) div 255) / $7FFFFF;
      end;
    wsfInt32:
      Result := 1.0 * PLongInt(FData + idx)^ / $7FFFFFFF;
    wsfInt64:
      Result := 1.0 * PInt64(FData + idx)^ / $7FFFFFFFFFFFFFFF;
    wsfFloat32:
      Result := PSingle(FData + idx)^;
    wsfFloat64:
      Result := PDouble(FData + idx)^;
  end;
end;

function TWavFile.GetValueAsInt(ASampleIndex, AChannelIndex: Integer): Int64;
var
  idx: Integer;
  d24: TData4 = (0, 0, 0, 0);
begin
  idx := ASampleIndex * FHeader.BlockAlign + AChannelIndex * FHeader.BitsPerSample div 8;
  case FSampleFormat of
    wsfInt8:
      Result := PByte(FData + idx)^;
    wsfInt16:
      Result := PSmallInt(FData + idx)^;
    wsfInt24:
      begin
        Move((FData + idx)^, d24[1], 3);
        Result := LongInt(d24) div 255;
      end;
    wsfInt32:
      Result := PLongInt(FData + idx)^;
    wsfInt64:
      Result := PInt64(FData + idx)^;
    wsfFloat32:
      Result := round(PSingle(FData + idx)^ * $7FFFFFFFFFFFFFFF);
    wsfFloat64:
      Result := round(PDouble(FData + idx)^ * $7FFFFFFFFFFFFFFF);
  end;
end;

function TWavFile.LoadFromStream(AStream: TStream; out AMsg: String): Boolean;
var
  RiffHdr: TWavRiffHeader;
  chunkHdr: TWavSubChunkHeader;
  s: String;
begin
  Result := false;
  AMsg := '';

  if FData <> nil then begin
    FreeMem(FData, FDataSize);
    FDataSize := 0;
  end;

  AStream.Position := 0;
  if AStream.Read(RiffHdr, SizeOf(RiffHdr)) <> SizeOf(RiffHdr) then begin
    AMsg := 'No valid RIFF format';
    exit;
  end;

  SetLength(s, 4);
  Move(RiffHdr.ID, s[1], 4);
  if s <> 'RIFF' then begin
    AMsg := 'No valid RIFF format.';
    exit;
  end;

  FChunkSize := RiffHdr.ChunkSize;

  Move(RiffHdr.RiffType, s[1], 4);
  if s <> 'WAVE' then begin
    AMsg := 'No valid WAV format.';
    exit;
  end;

  if AStream.Read(FHeader, SizeOf(Header)) <> SizeOf(Header) then begin
    AMsg := 'No valid WAV format.';
    exit;
  end;

  Move(FHeader.ID, s[1], 4);
  if s <> 'fmt ' then begin
    AMsg := 'No valid WAV format.';
    exit;
  end;

  if FHeader.SubChunkSize <> 16 then begin
    AMsg := 'No valid WAV format.';
    exit;
  end;

  if FHeader.AudioFormat in [1, 3] then begin
    while AStream.Position < AStream.Size do begin
      if AStream.Read(chunkHdr, SizeOf(chunkHdr)) <> SizeOf(chunkHdr) then begin
        AMsg := 'No valid WAV format.';
        Exit;
      end;
      SetLength(s, 4);
      Move(chunkHdr.ID, s[1], 4);
      if s = 'data' then begin
        FDataSize := chunkHdr.DataSize;
        GetMem(FData, FDataSize);
        if AStream.Read(FData^, FDataSize) <> FDataSize then begin
          AMsg := 'No valid WAV format';
          FreeMem(FData);
          FData := nil;
          FDataSize := 0;
          exit;
        end;
        break;
      end else
        AStream.Position := AStream.Position + chunkHdr.DataSize;
    end;
    FSampleFormat := GetSampleFormat;
  end else begin
    Result := true;
    AMsg := 'Cannot handle compressed data.';
  end;

  Result := true;
end;


end.

